/*
 * ipsec_alg 3DES cipher stubs
 *
 * Copyright (C) 2005 Michael Richardson <mcr@xelerance.com> 
 *
 * Adapted from ipsec_alg_aes.c by JuanJo Ciarlante <jjo-ipsec@mendoza.gov.ar>
 * 
 * ipsec_alg_aes.c,v 1.1.2.1 2003/11/21 18:12:23 jjo Exp
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 */
#include <linux/config.h>
#include <linux/version.h>

/*	
 *	special case: ipsec core modular with this static algo inside:
 *	must avoid MODULE magic for this file
 */
#if CONFIG_KLIPS_MODULE && CONFIG_KLIPS_ENC_1DES
#undef MODULE
#endif

#include <linux/module.h>
#include <linux/init.h>

#include <linux/kernel.h> /* printk() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/string.h>
#ifdef CONFIG_HW_C6010_ENC	// added by lily@amit.com.tw, for C6010 HW support -- 20060221
#include <net/c6010_sym.h>
#endif

/*	Low freeswan header coupling	*/
#include "openswan/ipsec_xform.h"
#include "openswan/ipsec_alg.h"
#include "crypto/des.h"
#include "openswan/ipsec_alg_1des.h"

static int debug_1des=0;
static int test_1des=0;
static int excl_1des=0;

#if defined(CONFIG_KLIPS_ENC_1DES_MODULE)
MODULE_AUTHOR("Michael Richardson <mcr@xelerance.com>");
MODULE_PARM(debug_1des, "i");
MODULE_PARM(test_des, "i");
MODULE_PARM(excl_des, "i");
#endif

static int _1des_set_key(struct ipsec_alg_enc *alg,
			 __u8 * key_e, const __u8 * key,
			 size_t keysize)
{
	int ret = 0;
	OneDES_context *ctx = (OneDES_context*)key_e;

	if(keysize != 192/8) {
	  return EINVAL;
	}
	
	des_set_key((des_cblock *)(key + DES_KEY_SZ*0), ctx->s1);
	
	if (debug_1des > 0)
		printk(KERN_DEBUG "klips_debug:_1des_set_key:"
				"ret=%d key_e=%p key=%p keysize=%ld\n",
                                ret, key_e, key, (unsigned long int) keysize);
	return ret;
}

static int _1des_cbc_encrypt(struct ipsec_alg_enc *alg,
			     __u8 * key_e,
			     __u8 * in,
			     int ilen, const __u8 * iv,
			     int encrypt)
{
#ifdef CONFIG_HW_C6010_ENC	// added by lily@amit.com.tw for C6010 support -- 20060221
	des_cblock miv;

	memcpy(&miv, iv, sizeof(miv));
	if(encrypt)
		C6010_DES_ENC(in, in, key_e, &miv, ilen);
	else
		C6010_DES_DEC(in, in, key_e, &miv, ilen);
#else
	OneDES_context *ctx=(OneDES_context*)key_e;
	des_cblock miv;

	memcpy(&miv, iv, sizeof(miv));

	if (debug_1des > 0)
		printk(KERN_DEBUG "klips_debug:_1des_cbc_encrypt:"
				"key_e=%p in=%p ilen=%d iv=%p encrypt=%d\n",
				key_e, in, ilen, iv, encrypt);

	des_cbc_encrypt((des_cblock *)in,
			     (des_cblock *)in,
			     ilen,
			     ctx->s1,
			     &miv, encrypt);
#endif
	return 1;
}

static struct ipsec_alg_enc ipsec_alg_1DES = {
	ixt_common: {	ixt_version:	IPSEC_ALG_VERSION,
			ixt_refcnt:	ATOMIC_INIT(0),
			ixt_name: 	"1des",
			ixt_blocksize:	ESP_1DES_CBC_BLK_LEN, 
			ixt_support: {
			  ias_exttype:	  IPSEC_ALG_TYPE_ENCRYPT,
			  ias_id: 	  ESP_DES,
			  ias_keyminbits: ESP_1DES_KEY_SZ*8,
			  ias_keymaxbits: ESP_1DES_KEY_SZ*8,
		},
	},
#if defined(MODULE_KLIPS_ENC_1DES_MODULE)
	ixt_module:	THIS_MODULE,
#endif
	ixt_e_keylen:	ESP_1DES_KEY_SZ*8,
	ixt_e_ctx_size:	sizeof(OneDES_context),
	ixt_e_set_key:	_1des_set_key,
	ixt_e_cbc_encrypt:_1des_cbc_encrypt,
};

#if defined(CONFIG_KLIPS_ENC_1DES_MODULE)
IPSEC_ALG_MODULE_INIT_MOD( ipsec_1des_init )
#else
IPSEC_ALG_MODULE_INIT_STATIC( ipsec_1des_init )
#endif
{
	int ret, test_ret;

	if (excl_1des) ipsec_alg_1DES.ixt_common.ixt_state |= IPSEC_ALG_ST_EXCL;
	ret=register_ipsec_alg_enc(&ipsec_alg_1DES);
	printk("ipsec_1des_init(alg_type=%d alg_id=%d name=%s): ret=%d\n", 
			ipsec_alg_1DES.ixt_common.ixt_support.ias_exttype, 
			ipsec_alg_1DES.ixt_common.ixt_support.ias_id, 
			ipsec_alg_1DES.ixt_common.ixt_name, 
			ret);
	if (ret==0 && test_1des) {
		test_ret=ipsec_alg_test(
				ipsec_alg_1DES.ixt_common.ixt_support.ias_exttype,
				ipsec_alg_1DES.ixt_common.ixt_support.ias_id, 
				test_1des);
		printk("ipsec_1des_init(alg_type=%d alg_id=%d): test_ret=%d\n", 
				ipsec_alg_1DES.ixt_common.ixt_support.ias_exttype, 
				ipsec_alg_1DES.ixt_common.ixt_support.ias_id, 
				test_ret);
	}
	return ret;
}

#if defined(CONFIG_KLIPS_ENC_1DES_MODULE)
IPSEC_ALG_MODULE_EXIT_MOD( ipsec_1des_fini )
#else
IPSEC_ALG_MODULE_EXIT_STATIC( ipsec_1des_fini )
#endif
{
	unregister_ipsec_alg_enc(&ipsec_alg_1DES);
	return;
}

/* Dual, because 3des code is 4-clause BSD licensed */
#ifdef MODULE_LICENSE
MODULE_LICENSE("Dual BSD/GPL");
#endif


